# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg

from networking_dpm.conf.types import NetworkAdapterMappingType


class MultiNetworkAdapterMappingOpt(cfg.MultiOpt):
    def __init__(self, name, **kwargs):
        """Option with DPM NetworkAdapterMappingType type

           Option with ``type`` :class:`NetworkAdapterMappingType`

        :param name: the option's name
        """
        super(MultiNetworkAdapterMappingOpt, self).__init__(
            name, item_type=NetworkAdapterMappingType(), **kwargs)



# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg

OBJECT_ID_REGEX = "[a-f0-9]{8}-([a-f0-9]{4}-){3}[a-f0-9]{12}"
PORT_REGEX = "[0,1]"
MAPPING_REGEX = "^[^:]+:" + OBJECT_ID_REGEX + "(:?|(:" + PORT_REGEX + ")?)$"


class NetworkAdapterMappingType(cfg.types.String):
    """Network adapter mapping type.

    Values are returned as tuple (net, adapter-id, port).
    Port defaults to '0' if non given.
    """
    def __init__(self, type_name='multi valued'):
        super(NetworkAdapterMappingType, self).__init__(
            type_name=type_name,
            regex=MAPPING_REGEX,
            ignore_case=True
        )

    def format_defaults(self, default, sample_default=None):
        multi = cfg.types.MultiString()
        return multi.format_defaults(default, sample_default)

    def __call__(self, value):
        val = super(NetworkAdapterMappingType, self).__call__(value)
        # No extra checking for None required here.
        # The regex ensures the format of the value in the super class.
        split_result = val.split(':')
        net = split_result[0]
        adapter_id = split_result[1].lower()
        port = (split_result[2] if len(split_result) == 3 and
                split_result[2] else "0")
        return net, adapter_id, port



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from zhmcclient import NotFound
from zhmcclient import NoUniqueMatch

"""Fake zhmcclient for testing

The fake zhmcclient provides an easy way of testing code that includes READ
calls to the real zhmcclient [1].

Usage:
Just override the real zhmcclient with this mock client in the test
case. As input parameter the fake zhmcclient takes a json dict that represents
the HMC. For an example, see the corresponding test case.

[1] https://github.com/zhmcclient/python-zhmcclient
"""


class _BaseObject(object):
    def __init__(self, object_json):
        self.properties = object_json

    def get_property(self, name):
        # Throws KeyError if not found
        return self.properties[name]


class _BaseManager(object):
    def __init__(self, type, json_object_list):
        self._objects = []
        for object in json_object_list:
            self._objects.append(type(object))

    def list(self):
        return self._objects

    def _get(self, object_id):
        """This is NOT an zhmclcient method, but used for test only"""
        return self.find(**{'object-id': object_id})

    def find(self, **kwargs):
        matches = self.findall(**kwargs)
        num_matches = len(matches)
        if num_matches == 0:
            raise NotFound
        elif num_matches > 1:
            raise NoUniqueMatch
        else:
            return matches[0]

    def findall(self, **kwargs):
        filter_criteria = kwargs.items()
        found = []
        for obj in self._objects:
            try:
                if all(obj.get_property(propname) == value
                       for (propname, value) in filter_criteria):
                    found.append(obj)
            except AttributeError:
                continue
        return found


class _NIC(_BaseObject):
    def __init__(self, nic_json):
        super(_NIC, self).__init__(nic_json)


class _Port(_BaseObject):
    def __init__(self, port_json):
        # Ports 'element-id' attribute is of type string, while a
        # vswitchs 'port' attribute is a int
        element_id = port_json.get('element-id')
        if (element_id is not None and
                not isinstance(element_id, str)):
            raise TypeError("Element-id attribute must be of type 'str', but "
                            "is of type '%s'." % type(element_id))
        super(_Port, self).__init__(port_json)


class _Adapter(_BaseObject):
    def __init__(self, adapter_json):
        self.ports = _PortManager(adapter_json.pop('ports', {}))
        super(_Adapter, self).__init__(adapter_json)


class _VSwitch(_BaseObject):
    def __init__(self, vswitch_json):
        self.nics = [_NIC(nic) for nic in vswitch_json.pop('nics', [])]
        # vswitch 'port' attribute is of type int, while an adapter ports
        # 'element-id' is a string
        port = vswitch_json.get('port')
        if port is not None and not isinstance(port, int):
            raise TypeError("Port attribute must be of type 'int', but is of "
                            "type '%s'." % type(port))
        super(_VSwitch, self).__init__(vswitch_json)

    def get_connected_nics(self):
        return self.nics


class _PortManager(_BaseManager):
    def __init__(self, port_list):
        super(_PortManager, self).__init__(_Port, port_list)


class _AdapterManager(_BaseManager):
    def __init__(self, adapter_list):
        super(_AdapterManager, self).__init__(_Adapter, adapter_list)


class _VSwitchManager(_BaseManager):
    def __init__(self, vswitch_list):
        super(_VSwitchManager, self).__init__(_VSwitch, vswitch_list)


class _CPC(_BaseObject):

    def __init__(self, cpc_json):
        self.vswitches = _VSwitchManager(cpc_json.pop('vswitches', {}))
        self.adapters = _AdapterManager(cpc_json.pop('adapters', {}))
        self.dpm_enabled = cpc_json.pop('dpm_enabled', True)
        super(_CPC, self).__init__(cpc_json)


class _CPCManager(_BaseManager):

    def __init__(self, cpc_list):
        super(_CPCManager, self).__init__(_CPC, cpc_list)


class _Client(object):
    def __init__(self, hmc_json):
        self.cpcs = _CPCManager(hmc_json['cpcs'])


class FakeZHMCClientError(Exception):
    pass


def _client_consistency_check(client):
    cpcs = client.cpcs.list()
    for cpc in cpcs:
        if not cpc.adapters or not cpc.adapters.list():
            # No adapters given - no consistency check
            continue
        # Validate that for each virtual switch object a corresponding
        # adapter exists, if adapters and virtual switches are given for a
        # cpc
        for vswitch in cpc.vswitches.list():
            uri = vswitch.get_property('backing-adapter-uri')
            try:
                # Fails with NotFound if adapter doesn't exist
                cpc.adapters.find(
                    **{'object-id': uri.replace('/api/adapters/', '')})
            except Exception:
                raise FakeZHMCClientError(
                    "No adapter object for virtual switch object %s found!"
                    % vswitch)


def get_client(hmc_json):
    """Get an instance of a faked zhmcclient

    :param hmc_json: A json dict or json string representing the hmc
    :return: faked zhmcclient client instance
    """
    # Test if input is a valid json
    if type(hmc_json) == dict:
        json.dumps(hmc_json)
    elif type(hmc_json) == str:
        hmc_json = json.loads(hmc_json)

    client = _Client(hmc_json)
    _client_consistency_check(client)
    return client


def get_cpc(test_json):
    """Get an instance of a faked zhmcclient cpc

    :param test_json: A json dict or json string representing the hmc
    :return: faked zhmcclient cpc instance
    """
    client = get_client(test_json)
    return client.cpcs.list()[0]



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_dpm.config import config as os_dpm_conf
from oslo_config import cfg

from networking_dpm.conf import config
from networking_dpm.tests import base


class TestConfig(base.BaseTestCase):
    def test_config(self):
        cfg.CONF.set_override('hmc', 'hmc-ip', 'dpm')
        cfg.CONF.set_override('hmc_username', 'username', 'dpm')
        cfg.CONF.set_override('hmc_password', 'password', 'dpm')
        cfg.CONF.set_override('cpc_object_id', 'oid', 'dpm')
        cfg.CONF.set_override('physical_network_adapter_mappings', 'mapping',
                              'dpm')

        self.assertEqual('hmc-ip', cfg.CONF.dpm.hmc)
        self.assertEqual('username', cfg.CONF.dpm.hmc_username)
        self.assertEqual('password', cfg.CONF.dpm.hmc_password)
        self.assertEqual('oid', cfg.CONF.dpm.cpc_object_id)
        self.assertEqual('mapping',
                         cfg.CONF.dpm.physical_network_adapter_mappings)

    def test_list_opts(self):
        expected = os_dpm_conf.COMMON_DPM_OPTS + config.dpm_opts
        # Result is a list of tuples (one tuple per config group)
        result = config.list_opts()

        # DPM Config options
        dpm_tuple = result[0]
        # Tuple consists of the groups name and a list of config_option objects
        self.assertEqual(os_dpm_conf.DPM_GROUP, dpm_tuple[0])
        config_opts = dpm_tuple[1]
        self.assertItemsEqual(expected, config_opts)



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import mock
from zhmcclient import ConnectionError
from zhmcclient import HTTPError

from networking_dpm.ml2 import dpm_neutron_agent as dpm_agt
from networking_dpm.ml2.dpm_neutron_agent import (PhysicalNetworkMapping as
                                                  dpm_map)
from networking_dpm.tests import base
from networking_dpm.tests.unit import fake_zhmcclient

from neutron.common import topics

OID_A = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
OID_B = "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
OID_C = "cccccccc-cccc-cccc-cccc-cccccccccccc"


class TestPhysnetMapping(base.BaseTestCase):

    def test_create_mapping(self):
        conf_mapping = ["physnet1:" + OID_A + ":0",
                        "physnet2:" + OID_B + ":1",
                        "physnet3:" + OID_C + ":0"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)
        expected = {'physnet1': ['vswitch-uuid-1'],
                    'physnet2': ['vswitch-uuid-2'],
                    'physnet3': ['vswitch-uuid-3']}

        adapters = [{'object-id': OID_A,
                     'type': 'osd', 'ports': [{'element-id': '0'}]},
                    {'object-id': OID_B,
                     'type': 'osd', 'ports': [{'element-id': '1'}]},
                    {'object-id': OID_C,
                     'type': 'hipersockets', 'ports': [{'element-id': '0'}]}]
        vswitches = [
            {"backing-adapter-uri": "/api/adapters/" + OID_A,
             "object-id": "vswitch-uuid-1",
             "port": 0},
            {"backing-adapter-uri": "/api/adapters/" + OID_B,
             "object-id": "vswitch-uuid-2",
             "port": 1},
            {"backing-adapter-uri": "/api/adapters/" + OID_C,
             "object-id": "vswitch-uuid-3",
             "port": 0}]
        hmc = {"cpcs": [{"object-id": "cpcpid", "vswitches": vswitches,
                         "adapters": adapters}]}
        cpc = fake_zhmcclient.get_cpc(hmc)
        mapping = dpm_map.create_mapping(cpc)
        actual = mapping.get_mapping()
        self.assertItemsEqual(expected['physnet1'], actual['physnet1'])
        self.assertItemsEqual(expected['physnet2'], actual['physnet2'])
        self.assertItemsEqual(expected['physnet3'], actual['physnet3'])
        vswitch_list = mapping.get_all_vswitches()
        vswitch_ids = [vswitch.get_property('object-id')
                       for vswitch in vswitch_list]

        self.assertIn("vswitch-uuid-1", vswitch_ids)
        self.assertIn("vswitch-uuid-2", vswitch_ids)
        self.assertIn("vswitch-uuid-3", vswitch_ids)

    def test_create_mapping_invalid_adapter_type(self):
        conf_mapping = ["physnet1:" + OID_A + ":0"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)
        hmc = {"cpcs": [{
            "object-id": "cpcpid",
            "adapters": [{'object-id': 'uuid-1', 'type': 'bad_type'}]}]}

        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertRaises(SystemExit,
                          dpm_map.create_mapping, cpc)

    def test_create_mapping_adapter_not_exists(self):
        conf_mapping = ["physnet1:" + OID_A + ":0"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)
        hmc = {"cpcs": [{
            "object-id": "cpcpid",
            "adapters": [{'object-id': 'other_adapter', 'type': 'osd'}]}]}

        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertRaises(SystemExit,
                          dpm_map.create_mapping, cpc)

    def test_create_mapping_adapter_port_not_exists(self):
        conf_mapping = ["physnet1:" + OID_A + ":1"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)
        hmc = {"cpcs": [{
            "object-id": "cpcpid",
            "adapters": [{'object-id': OID_A, 'type': 'osd',
                          'ports': []}]}]}

        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertRaises(SystemExit,
                          dpm_map.create_mapping, cpc)

    def test_create_mapping_vswitch_not_exists(self):
        conf_mapping = ["physnet1:" + OID_A + ":0"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)
        hmc = {"cpcs": [{
            "object-id": "cpcpid",
            "adapters": [{'object-id': OID_A, 'type': 'osd'}]}]}

        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertRaises(SystemExit,
                          dpm_map.create_mapping, cpc)

    def test_create_mapping_multiple_adapters_per_physnet(self):
        conf_mapping = ["physnet1:" + OID_A + ":0",
                        "physnet1:" + OID_B + ":0"]
        self.flags(group="dpm", physical_network_adapter_mappings=conf_mapping)

        adapters = [{'object-id': OID_A, 'type': 'osd'},
                    {'object-id': OID_B, 'type': 'osd'}]
        hmc = {"cpcs": [{"object-id": "cpcpid", "vswitches": [
            {"backing-adapter-uri": "/api/adapters/" + OID_A,
             "object-id": "vswitch-uuid-1",
             "port": 0},
            {"backing-adapter-uri": "/api/adapters/" + OID_B,
             "object-id": "vswitch-uuid-3",
             "port": 0}], "adapters": adapters}]}
        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertRaises(SystemExit,
                          dpm_map.create_mapping,
                          cpc)


class TestDPMRPCCallbacks(base.BaseTestCase):
    def setUp(self):
        super(TestDPMRPCCallbacks, self).setUp()

        agent = mock.Mock()
        agent.mgr = mock.Mock()
        agent.mgr.interface_mappings = 'mappings'
        self.rpc = dpm_agt.DPMRPCCallBack(mock.Mock(), agent, mock.Mock())

    def test_port_update(self):
        port = {'id': 'port-id123', 'mac_address': 'mac1'}
        self.rpc.port_update(context=None, port=port)
        self.assertEqual(set(['mac1']), self.rpc.updated_devices)


class TestDPMManager(base.BaseTestCase):
    def setUp(self):
        super(TestDPMManager, self).setUp()
        self.mgr = dpm_agt.DPMManager({}, mock.Mock(), [])

    def test__managed_by_agent(self):
        self.flags(host='foo-host')
        valid_mac_str = "mac=00:11:22:33:44:55:66"
        # Host and mac part of description
        self.assertTrue(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': 'OpenStackcontains-foo-host-in-description' +
                            valid_mac_str})))

        # Unicode
        self.assertTrue(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': u"OpenStack foo-host mac=00:11:22:33:44:55:66"})))

        # host missing
        self.assertFalse(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': 'OpenStacknot-managed' + valid_mac_str})))

        # mac missing
        self.assertFalse(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': 'OpenStack foo-host'})))

        # prefix missing
        self.assertFalse(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': 'foo-host' + valid_mac_str})))

        # No host match
        self.assertFalse(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': 'not-managed' + valid_mac_str})))

        # No description
        self.assertFalse(self.mgr._managed_by_agent(fake_zhmcclient._NIC(
            {'description': ''})))

    def test_get_all_devices(self):
        hmc = {"cpcs": [{"object-id": "cpcpid", "vswitches": [
            {"backing-adapter-uri": "/api/adapters/uuid-1",
             "object-id": "vswitch-uuid-1",
             "port": 0,
             "nics": [{"description": "foomac=00:00:00:00:00:11bar"},
                      {"description": "foomac=00:00:00:00:00:33bar"}]},
            {"backing-adapter-uri": "/api/adapters/uuid-2",
             "object-id": "vswitch-uuid-2",
             "port": 1,
             "nics": [{"description": "foomac=00:00:00:00:00:22bar"}]},
            {"backing-adapter-uri": "/api/adapters/uuid-3",
             "object-id": "vswitch-uuid-3",
             "port": 0},
            {"backing-adapter-uri": "/api/adapters/not-configured",
             "object-id": "not-configured",
             "port": 1,
             "nics": [{"description": "not-configured"}]},
        ]}]}

        cpc = self.mgr.cpc = fake_zhmcclient.get_cpc(hmc)

        self.mgr.vswitches = [cpc.vswitches._get('vswitch-uuid-1'),
                              cpc.vswitches._get('vswitch-uuid-2'),
                              cpc.vswitches._get('vswitch-uuid-3')]
        with mock.patch.object(self.mgr, '_managed_by_agent',
                               return_value=True) as is_uuid:
            devices = self.mgr.get_all_devices()
            self.assertEqual(3, is_uuid.call_count)
        expected = ['00:00:00:00:00:11', '00:00:00:00:00:22',
                    '00:00:00:00:00:33']
        self.assertItemsEqual(expected, devices)

    def test_get_all_devices_mac_not_present(self):
        hmc = {"cpcs": [{"object-id": "cpcpid", "vswitches": [
            {"backing-adapter-uri": "/api/adapters/uuid-1",
             "object-id": "vswitch-uuid-1",
             "port": 0,
             "nics": [{"description": "OpenStack foo"}]},
        ]}]}

        cpc = self.mgr.cpc = fake_zhmcclient.get_cpc(hmc)

        self.mgr.vswitches = [cpc.vswitches._get('vswitch-uuid-1')]
        devices = self.mgr.get_all_devices()
        expected = []
        self.assertItemsEqual(expected, devices)

    def test_get_all_devices_deleted_concurrently(self):
        hmc = {"cpcs": [{"object-id": "cpcpid", "vswitches": [
            {"backing-adapter-uri": "/api/adapters/uuid-1",
             "object-id": "vswitch-uuid-1",
             "port": 0,
             "nics": [{"name": "port-id-1"}]},
        ]}]}
        cpc = self.mgr.cpc = fake_zhmcclient.get_cpc(hmc)
        self.mgr.vswitches = [cpc.vswitches._get('vswitch-uuid-1')]
        with mock.patch.object(fake_zhmcclient._NIC, 'get_property',
                               side_effect=HTTPError(mock.Mock())):
            devices = self.mgr.get_all_devices()
            self.assertEqual(0, len(devices))

    def _test_get_all_devices_vswitch_error(self, error):
        hmc = {"cpcs": [{"object-id": "cpcpid"}]}

        self.mgr.cpc = fake_zhmcclient.get_cpc(hmc)
        vswitch = mock.Mock()
        self.mgr.vswitches = [vswitch]
        with mock.patch.object(vswitch, 'get_connected_nics',
                               side_effect=error):
            self.mgr.get_all_devices()

    @mock.patch.object(sys, 'exit')
    def test_get_all_devices_vswitch_http_error_ignore(self, m_exit):
        self._test_get_all_devices_vswitch_error(HTTPError({}))
        self.assertFalse(m_exit.called)

    @mock.patch.object(sys, 'exit')
    def test_get_all_devices_vswitch_http_404(self, m_exit):
        http_error = HTTPError({'http-status': 404})
        self._test_get_all_devices_vswitch_error(http_error)
        m_exit.assert_called_once_with(1)

    def test_get_all_devices_connection_error(self):
        vswitch_bad = mock.Mock()
        vswitch_bad.get_connected_nics.side_effect = ConnectionError("foo")
        self.mgr.vswitches = [vswitch_bad]

        devices = self.mgr.get_all_devices()
        self.assertEqual(set(), devices)

    def test_get_agent_configurations(self):
        self.mgr.physnet_map = 'foo'
        expected = {'adapter_mappings': 'foo'}
        self.assertEqual(expected, self.mgr.get_agent_configurations())

    def test_get_agent_id(self):
        self.flags(host='foo')
        self.assertEqual("dpm-foo", self.mgr.get_agent_id())

    def test_get_extension_driver_type(self):
        self.assertEqual('dpm', self.mgr.get_extension_driver_type())

    def test_get_rpc_callbacks(self):
        context = mock.Mock()
        agent = mock.Mock()
        sg_agent = mock.Mock()
        obj = self.mgr.get_rpc_callbacks(context, agent, sg_agent)
        self.assertIsInstance(obj, dpm_agt.DPMRPCCallBack)

    def test_get_rpc_consumers(self):
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        self.assertEqual(consumers, self.mgr.get_rpc_consumers())

    def test_plug_interface(self):
        self.assertTrue(self.mgr.plug_interface('network_id',
                                                'network_segment', 'mac1',
                                                'device_owner'))

    def test__extract_mac(self):
        desc = "Openstack:mac=00:11:22:33:44:55,host_id=foo"
        nic = fake_zhmcclient._NIC({'description': desc})
        self.assertEqual("00:11:22:33:44:55", self.mgr._extract_mac(nic))


class TestDPMMain(base.BaseTestCase):

    def test__validate_firewall_driver(self):
        valid_drivers = ['neutron.agent.firewall.NoopFirewallDriver',
                         'noop']
        for driver in valid_drivers:
            self.flags(group='SECURITYGROUP', firewall_driver=driver)
            dpm_agt._validate_firewall_driver()

    @mock.patch.object(sys, 'exit')
    def test__validate_firewall_driver_invalid(self, mock_exit):
        self.flags(group='SECURITYGROUP', firewall_driver='foo')
        dpm_agt._validate_firewall_driver()
        mock_exit.assert_called_with(1)

    def test__get_cpc_found(self):
        hmc = {"cpcs": [{"object-id": "oid1"}]}
        client = fake_zhmcclient.get_client(hmc)
        cpc = dpm_agt._get_cpc(client, "oid1")
        self.assertEqual("oid1", cpc.get_property("object-id"))

    def test__get_cpc_not_dpm(self):
        hmc = {"cpcs": [{"object-id": "oid1", "dpm_enabled": False}]}
        client = fake_zhmcclient.get_client(hmc)

        with mock.patch.object(sys, 'exit') as m_sys:
            dpm_agt._get_cpc(client, 'oid1')
            m_sys.assert_called_with(1)

    def test__get_cpc_not_found(self):
        hmc = {"cpcs": [{"object-id": "oid1"}]}
        client = fake_zhmcclient.get_client(hmc)
        with mock.patch.object(sys, 'exit') as m_sys:
            dpm_agt._get_cpc(client, 'not-found')
            m_sys.assert_called_with(1)

    @mock.patch.object(dpm_agt.common_config, 'init')
    @mock.patch.object(dpm_agt, '_get_cpc')
    @mock.patch.object(dpm_agt.PhysicalNetworkMapping, 'create_mapping')
    @mock.patch.object(dpm_agt.service, 'launch')
    def test_main(self, mock_launch, mock_cm, mock_get_cpcp, mock_init):
        # This test also discovers when config options have changed e.g. in
        # os-dpm
        dpm_agt.main()
        mock_launch.assert_called_once()



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
from requests.packages import urllib3
import sys

from oslo_log import log as logging
import oslo_messaging
from oslo_service import service
import zhmcclient

from networking_dpm.conf import config
from networking_dpm.ml2 import constants as const
from networking_dpm.ml2.mech_dpm import AGENT_TYPE_DPM

from neutron._i18n import _LE
from neutron._i18n import _LI
from neutron._i18n import _LW
from neutron.api.rpc.handlers import securitygroups_rpc as sg_rpc
from neutron.common import config as common_config
from neutron.common import topics
from neutron.plugins.ml2.drivers.agent import _agent_manager_base as amb
from neutron.plugins.ml2.drivers.agent import _common_agent as ca

CONF = config.cfg.CONF
LOG = logging.getLogger(__name__)

NIC_OWNER_OPENSTACK = 'OpenStack'
DPM_AGENT_BINARY = "neutron-dpm-agent"
EXTENSION_DRIVER_TYPE = 'dpm'
ADAPTER_URI = "/api/adapters/"

# TODO(andreas_s): Suppressing the following warning thrown by urllib3 used by
# zhmcclient: "InsecureRequestWarning: Unverified HTTPS request is being made.
# Adding certificate verification is strongly advised. See:
# https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings"
urllib3.disable_warnings()


class PhysicalNetworkMapping(object):
    """Mapping of physical networks to vswitches

    Used by
    * the regular notifications sent to the Neutron server
    * the polling against the HMC for new NICs
    """

    def __init__(self, cpc):
        self._cpc = cpc
        self._vswitches = []
        self._physnet_mapping = {}

    def _get_vswitch(self, adapter_id, port):
        try:
            # TODO(andreas_s): Optimize in zhmcclient - For 'find' the
            # whole list of items is retrieved
            return self._cpc.vswitches.find(**{
                'backing-adapter-uri': ADAPTER_URI + adapter_id,
                # 'port' attribute of vswitch is of type int. But 'element-id'
                # attribute of an adapter port is still of type str.
                'port': int(port)
            })
        except zhmcclient.NotFound:
            LOG.error(
                _LE("No vswitch object for adapter/port combination "
                    "%(adapt)s/%(port)s found. Agent terminated!"),
                {'adapt': adapter_id, 'port': port}
            )
            sys.exit(1)

    def _add_vswitch(self, physnet, adapter_id, port):
        vswitch = self._get_vswitch(adapter_id, port)
        self._vswitches.append(vswitch)
        if self._physnet_mapping.get(physnet):
            # TODO(andreas_s): Lift this restriction
            LOG.error(_LE("Multiple vswitches for physical network "
                          "%(net)s defined but only a single vswitch "
                          "definition per physical network is supported."
                          "Agent terminated!"),
                      {'net': physnet})
            sys.exit(1)
        self._physnet_mapping[physnet] = [vswitch.get_property('object-id')]

    def _get_adapter(self, adapter_id):
        try:
            # TODO(andreas_s): Optimize in zhmcclient - For 'find' the
            # whole list of items is retrieved
            return self._cpc.adapters.find(**{'object-id': adapter_id})
        except zhmcclient.NotFound:
            LOG.error(_LE("Configured adapter %s could not be "
                          "found. Please update the agent "
                          "configuration. Agent terminated!"),
                      adapter_id)
            sys.exit(1)

    @staticmethod
    def _validate_adapter_type(adapter):
        adapt_type = adapter.get_property('type')
        if adapt_type not in ['osd', 'hipersockets']:
            LOG.error(_LE("Configured adapter %s is not an OSA "
                          "or a hipersockets adapter. Please update "
                          "the agent configuration. Agent "
                          "terminated!"), adapter)
            sys.exit(1)

    @staticmethod
    def _validate_adapter_port(adapter, port):
        try:
            # TODO(andreas_s): Optimize in zhmcclient - For 'find' the
            # whole list of items is retrieved
            adapter.ports.find(**{'element-id': port})
        except zhmcclient.NotFound:
            LOG.error(_LE("Configured port %(port)s for adapter "
                          "%(adapt)s does not exist. Please update "
                          "the agent configuration. Agent "
                          "terminated!"), {'adapt': adapter,
                                           'port': port})
            sys.exit(1)

    @staticmethod
    def create_mapping(cpc):
        mapping = PhysicalNetworkMapping(cpc)

        for entry in CONF.dpm.physical_network_adapter_mappings:
            # Safe as config option ensures validity of the tuple
            net, adapter_uuid, port = entry

            adapter = mapping._get_adapter(adapter_uuid)
            mapping._validate_adapter_type(adapter)
            mapping._validate_adapter_port(adapter, port)
            mapping._add_vswitch(net, adapter_uuid, port)
        return mapping

    def get_all_vswitches(self):
        """Get a list of all vswitch objects

        :return : list of zhmcclient vswitch objects
        """
        return self._vswitches

    def get_mapping(self):
        """Get the physnet/vswitch mapping for reports to the Neutron server

        :return : physical network / list of vswitch_ids mapping
        :rtype : dict of lists of strings
        """
        return self._physnet_mapping


class DPMRPCCallBack(sg_rpc.SecurityGroupAgentRpcCallbackMixin,
                     amb.CommonAgentManagerRpcCallBackBase):
    target = oslo_messaging.Target(version='1.4')

    def port_update(self, context, **kwargs):
        port = kwargs['port']
        LOG.debug("port_update received for port %s ", port)
        mac = port['mac_address']
        # Put the device name in the updated_devices set.
        # Do not store port details, as if they're used for processing
        # notifications there is no guarantee the notifications are
        # processed in the same order as the relevant API requests.
        self.updated_devices.add(mac)


class DPMManager(amb.CommonAgentManagerBase):
    def __init__(self, physnet_mapping, cpc, vswitches):
        self.physnet_map = physnet_mapping
        self.vswitches = vswitches
        self.cpc = cpc

    def ensure_port_admin_state(self, device, admin_state_up):
        # Setting a port up/down is not supported by Neutron DPM agent. Ports
        # are always up.
        pass

    def get_agent_configurations(self):
        return {'adapter_mappings': self.physnet_map}

    def get_agent_id(self):
        return 'dpm-%s' % CONF.host

    def get_devices_modified_timestamps(self, devices):
        # TODO(andreas_s): this should be implemented to detect
        # rapid Nova instance rebuilds.
        return {}

    @staticmethod
    def _managed_by_agent(nic):
        """Verifies if a NIC is supposed to be managed by this agent

        On NIC creation, Nova adds the following to the NICs description
        * Identifier 'OpenStack' at the beginning
        * Neutron ports MAC 'mac=xx:xx:xx:xx:xx:xx'
        * Host identifier of the CPCSubset 'host-id=foo'

        This method is able determine along those parameters if a NIC
        object should be managed by this agent or not.

        :param nic: The nic that should be checked
        :type nic: zhmcclient._NIC
        :return: True if NIC is supposed to be managed by this agent
        :rtype: bool
        """
        # Nova sets the Name of the NIC object to the UUID of the Neutron port
        description = nic.get_property('description')
        if not description.startswith(NIC_OWNER_OPENSTACK):
            LOG.debug("NIC %(nic)s seems not managed by OpenStack, as its "
                      "description %(desc)s does not start with the NIC OWNER "
                      "identifier %(noo)s!",
                      {'nic': nic, 'desc': description,
                       'noo': NIC_OWNER_OPENSTACK})
            return False

        # Nova adds the host identifier to the NICs description attribute
        if CONF.host not in nic.get_property('description'):
            LOG.debug("NIC %(nic)s not managed by this host %(host)s. "
                      "Skipping.", {'nic': nic, 'host': CONF.host})
            return False

        # check if mac is present
        if not DPMManager._extract_mac(nic):
            LOG.debug("Description of NIC %s does not contain a valid mac "
                      "address. Therefore it seems not to be managed by this"
                      "Neutron agent. Skipping!", nic)
            return False
        return True

    @staticmethod
    def _extract_mac(nic):
        description = nic.get_property('description')
        mac_regex = 'mac=(([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2}))'
        match = re.search(mac_regex, description)
        if match:
            return match.group(1)
        # TODO(andreas_s): throw exception

    def _filter_agent_managed_nic_macs(self, nics):
        macs = set()
        for nic in nics:
            try:
                if self._managed_by_agent(nic):
                    macs.add(self._extract_mac(nic))
            except zhmcclient.HTTPError as http_error:
                if (http_error.http_status ==
                        const.HTTP_STATUS_NOT_FOUND):
                    LOG.debug("NIC %s got deleted concurrently."
                              "Continuing...", nic)
                else:
                    LOG.warning(_LW("NIC %(nic)s will be reported as "
                                    "'DOWN' due to error: %(err)s"),
                                {'nic': nic, 'err': http_error})
        return macs

    def get_all_devices(self):
        """Getting all NICs that are managed by this agent

        :return: List of Neutron port MACs for which NICs exist
        """
        devices = set()

        for vswitch in self.vswitches:
            try:
                nics = vswitch.get_connected_nics()
                devices = devices.union(
                    self._filter_agent_managed_nic_macs(nics))
            except zhmcclient.ConnectionError as con_err:
                LOG.error(_LE(
                    "%(message)s, %(details)s. Lost connection to HMC of "
                    "CPC %(cpc)s. All NICs of this CPC and its corresponding "
                    "Neutron ports wil be reported as 'DOWN'."),
                    {"cpc": self.cpc, "message": con_err,
                     "details": con_err.details})

            except zhmcclient.HTTPError as http_error:
                if http_error.http_status == const.HTTP_STATUS_NOT_FOUND:
                    LOG.error(_LE(
                        "An unrecoverable error occurred: %(err)s "
                        "DPM vSwitch object %(vswitch)s is "
                        "not available anymore. This can happen if "
                        "the corresponding adapter got removed "
                        "from the system or the corresponding "
                        "hipersockets network got deleted. Please "
                        "check the physical_network_adapter_mappings "
                        "configuration and start the "
                        "agent again. Agent terminated!"),
                        {'vswitch': vswitch, 'err': http_error})
                    # Need to exit with sys.exit, as calling code catches
                    # exceptions
                    sys.exit(1)

                LOG.warning(_LW("An error occurred while retrieving the "
                                "connected nics of vswitch %(vswitch)s: "
                                "%(err)s. All NICs of this vswitch will "
                                "be reported as 'DOWN'."),
                            {'vswitch': vswitch, 'err': http_error})
        return devices

    def get_extension_driver_type(self):
        return EXTENSION_DRIVER_TYPE

    def get_rpc_callbacks(self, context, agent, sg_agent):
        return DPMRPCCallBack(context, agent, sg_agent)

    def get_rpc_consumers(self):
        # TODO(andreas_s): remove security group update?
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        return consumers

    def plug_interface(self, network_id, network_segment, device,
                       device_owner):
        return True

    def setup_arp_spoofing_protection(self, device, device_details):
        # Arp spoofing protection is not supported by Neutron DPM agent
        pass

    def delete_arp_spoofing_protection(self, devices):
        # Arp spoofing protection is not supported by Neutron DPM agent
        pass

    def delete_unreferenced_arp_protection(self, current_devices):
        # Arp spoofing protection is not supported by Neutron DPM agent
        pass


def _validate_firewall_driver():
    fw_driver = CONF.SECURITYGROUP.firewall_driver
    supported_fw_drivers = ['neutron.agent.firewall.NoopFirewallDriver',
                            'noop']
    if fw_driver not in supported_fw_drivers:
        LOG.error(_LE('Unsupported configuration option for "SECURITYGROUP.'
                      'firewall_driver"! Only the NoopFirewallDriver is '
                      'supported by DPM agent, but "%s" is configured. '
                      'Set the firewall_driver to "noop" and start the '
                      'agent again. Agent terminated!'),
                  fw_driver)
        sys.exit(1)


def _get_cpc(client, cpc_oid):
    try:
        cpc = client.cpcs.find(**{'object-id': cpc_oid})
        if cpc.dpm_enabled:
            return cpc
        LOG.error(_LE("CPC %s not in DPM mode.") % cpc_oid)
    except zhmcclient.NotFound:
        LOG.error(_LE("Could not find CPC with object-id %s") % cpc_oid)
    sys.exit(1)


def setup_logging():
    """Sets up the logging options."""

    # We use the oslo.log default log levels but set the log level for
    # zhmcclient library to warning
    logging.set_defaults(default_log_levels=logging.get_default_log_levels() +
                         ["zhmcclient=WARNING"])
    logging.setup(CONF, 'neutron')
    LOG.info(_LI("Logging enabled!"))
    LOG.info(_LI("%(prog)s"), {'prog': sys.argv[0]})
    LOG.debug("command line: %s", " ".join(sys.argv))


def main():
    common_config.init(sys.argv[1:])
    setup_logging()

    hmc = CONF.dpm.hmc
    userid = CONF.dpm.hmc_username
    password = CONF.dpm.hmc_password
    cpc_oid = CONF.dpm.cpc_object_id

    session = zhmcclient.Session(hmc, userid, password)
    client = zhmcclient.Client(session)
    cpc = _get_cpc(client, cpc_oid)

    physnet_vswitch_map = PhysicalNetworkMapping.create_mapping(cpc)
    manager = DPMManager(physnet_vswitch_map.get_mapping(), cpc,
                         physnet_vswitch_map.get_all_vswitches())

    polling_interval = CONF.AGENT.polling_interval
    quitting_rpc_timeout = CONF.AGENT.quitting_rpc_timeout
    agent = ca.CommonAgentLoop(manager, polling_interval,
                               quitting_rpc_timeout,
                               AGENT_TYPE_DPM,
                               DPM_AGENT_BINARY)
    LOG.info(_LI("Agent initialized successfully, now running... "))
    service.launch(CONF, agent).wait()



#!/usr/bin/env python
# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ConfigParser
import re
import requests.packages.urllib3
from socket import gethostname
import time
import zhmcclient


from keystoneauth1 import identity
from keystoneauth1 import session
from neutronclient.v2_0 import client

TEST_PARTITION_NAME = 'networking-dpm-test'

# TODO(andreas_s): Implement using taskflow?

# OpenStack Authentication
os_user = 'admin'
os_password = 'openstack'
project_name = 'demo'
project_domain_id = 'default'
user_domain_id = 'default'
auth_url = 'http://localhost:5000/v3'
auth = identity.Password(auth_url=auth_url,
                         username=os_user,
                         password=os_password,
                         project_name=project_name,
                         project_domain_id=project_domain_id,
                         user_domain_id=user_domain_id)
sess = session.Session(auth=auth)
neutron = client.Client(session=sess)

# Parse neutron plugin conf
plugin_conf = ConfigParser.ConfigParser()
plugin_conf.read("/etc/neutron/plugins/ml2/ml2_conf.ini")
zhmc = plugin_conf.get('dpm', 'hmc')
userid = plugin_conf.get('dpm', 'hmc_username')
password = plugin_conf.get('dpm', 'hmc_password')
cpc_oid = plugin_conf.get('dpm', 'cpc_object_id')

adapter_mappings = plugin_conf.get('dpm', 'physical_adapter_mappings')
regex = re.search('public:(.{36}):([01]?)', adapter_mappings)
adapter_id = regex.group(1)
port_element_id = int(regex.group(2) or 0)


# Parse neutron.conf
plugin_conf = ConfigParser.ConfigParser()
plugin_conf.read("/etc/neutron/neutron.conf")

# Set host to cfg.CONF.DEFAULT.host if possible, default to hostname
host = None
try:
    host = plugin_conf.get('DEFAULT', 'host')
except ConfigParser.NoOptionError:
    host = gethostname()


# DPM Authentication
requests.packages.urllib3.disable_warnings()
session = zhmcclient.Session(zhmc, userid, password)
client = zhmcclient.Client(session)


cpc = client.cpcs.find(**{"object-id": cpc_oid})
print("Working on cpc %s" % cpc)

# Cleaning up partitions from previous test runs
for partition in cpc.partitions.findall(name=TEST_PARTITION_NAME):
    print("Cleaning up DPM partitions")
    print("  " + str(partition))
    partition.stop()
    partition.delete()

# Get virtual switch uri
vswitch = cpc.vswitches.find(**{
    'backing-adapter-uri': '/api/adapters/' + adapter_id,
    'port': port_element_id})
vswitch.pull_full_properties()
vswitch_id = vswitch.get_property('object-id')
vswitch_uri = vswitch.get_property('object-uri')

# Create OpenStack Port
networks = neutron.list_networks(name='private')
net = networks['networks'][0]
net_id = net['id']
prov_net_type = net['provider:network_type']
vlan_id = net['provider:segmentation_id']

body = {"port": {"network_id": net_id,
                 "binding:host_id": "vagrant-ubuntu-trusty-64"}}
port = neutron.create_port(body=body)
port_id = port['port']['id']
mac = port['port']['mac_address']
print("Created Neutron port with id %s" % str(port_id))

assert("DOWN" == port['port']['status'])

# Create Partition
properties = {
    'name': TEST_PARTITION_NAME,
    'description': 'Original partition description.',
    'cp-processors': 1,
    'initial-memory': 512,
    'maximum-memory': 512,
    'processor-mode': 'shared',
    'boot-device': 'test-operating-system'
}
partition = cpc.partitions.create(properties)
print("DPM partition created %s" % str(partition))

# Create the vnic and set the NICs properties like nova would do it
# nic.description starts with 'OpenStack'
# nic.description contains mac=<mac-of-neutron-port>
# nic.description contains host identifier
nic = partition.nics.create({"virtual-switch-uri": vswitch_uri,
                             "name": "OpenStack foo",
                             "description": "OpenStack mac=" + mac + " Host"
                                            + host})
print("DPM NIC created %s" % str(nic))

# sleeping to give the agent time to do the work
time.sleep(6)


def get_neutron_port(port_id):
    ports = neutron.list_ports(id=port_id)
    return ports['ports'][0]

port = get_neutron_port(port_id)

print(port)
# Do assertions
assert("ACTIVE" == get_neutron_port(port_id)['status'])
print("Neutron port status changed to 'ACTIVE'")
assert("dpm_vswitch" == port['binding:vif_type'])
print("Neutron port vif type set to 'dpm_vswitch'")
assert(vswitch_id == port['binding:vif_details']['object_id'])
print("Neutron port vif details object_id set to " + vswitch_id)
if prov_net_type == 'vlan':
    assert(vlan_id == port['binding:vif_details']['vlan'])
    print("VLAN ID set to " + str(vlan_id))
    assert('inband' == port['binding:vif_details']['vlan_mode'])
    print("vlan_mode set to 'inband'")

# Remove NIC from partition again
print("Removing DPM NIC again")
nic.delete()


# sleeping to give the agent time to do the work
time.sleep(6)

assert("DOWN" == get_neutron_port(port_id)['status'])
print("Neutron port status went to 'DOWN' again")

# Deleting Neutron port
neutron.delete_port(port_id)
print("Deleted Neutron port")

partition.delete()
print("Deleted DPM partition")
print("Success!")



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_dpm.config import config as os_dpm_conf
from oslo_config import cfg

from networking_dpm.conf.cfg import MultiNetworkAdapterMappingOpt

# Update help text of DPM group with networking-dpm specifics
os_dpm_conf.DPM_GROUP.help += """

DPM config options for the Neutron agent on the compute node (one agent
instance for each OpenStack hypervisor host) specify the target CPC, the HMC
managing it, and the OpenStack physical networks for the OpenStack hypervisor
host and their backing network adapters and ports in the target CPC."""

os_dpm_conf.register_opts()

mapping_example = ["physnet1:12345678-1234-1234-1234-123456789a",
                   "physnet2:12345678-1234-1234-1234-123456789b:1",
                   "physnet3:12345678-1234-1234-1234-123456789c:0"]

# TODO(andreas_s): Neutron does not make use of required=True, therefore
# the Neutron test base class tests fail when enabled
dpm_opts = [
    MultiNetworkAdapterMappingOpt(
        'physical_network_adapter_mappings',
        sample_default=mapping_example,
        help="""
The OpenStack physical networks that can be used by this OpenStack hypervisor
host, and their backing network adapters and ports in the target CPC.

This is a multi-line option. Each instance (line) of the option defines one
physical network for use by this OpenStack hypervisor host, and the network
adapter and port that is used for that physical network, using this syntax:

```
    <physical-network-name>:<adapter-object-id>[:<port-element-id>]
```

* `<physical-network-name>` is the name of the OpenStack physical network.
* `<adapter-object-id>` is the object-id of the network adapter in the target
  CPC that is used for this physical network.
* `<port-element-id>` is the element-id of the port on that network adapter.
  It is optional and defaults to 0.

The instances (lines) of this option for a particular Neutron agent

* must not specify the same physical network more than once, and
* must not specify the same adapter and port more than once, and
* must have all of their physical networks specified in the
  corresponding `*mappings` config option of the Neutron L2 agent service
  on all network nodes, and
* must have all of their physical networks specified in the
  `ml2.network_vlan_ranges` config option of the Neutron server, if vlan
  self service networks should be used.
""")


]

cfg.CONF.register_opts(dpm_opts, os_dpm_conf.DPM_GROUP)


def list_opts():
    return [
        (os_dpm_conf.DPM_GROUP, dpm_opts + os_dpm_conf.COMMON_DPM_OPTS),
    ]



# Copyright (c) 2016 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_dpm.ml2 import mech_dpm

from neutron.tests.unit.plugins.ml2 import _test_mech_agent as base


class DPMMechanismBaseTestCase(base.AgentMechanismBaseTestCase):
    VIF_TYPE = mech_dpm.VIF_TYPE_DPM_VSWITCH
    CAP_PORT_FILTER = False
    AGENT_TYPE = mech_dpm.AGENT_TYPE_DPM

    GOOD_MAPPINGS = {'fake_physical_network': ['fake_vswitch1',
                                               'fake_vswitch2']}
    GOOD_CONFIGS = {'adapter_mappings': GOOD_MAPPINGS}
    NO_VSWITCH_CONFIGS = {'adapter_mappings': {'fake_physical_network': []}}

    BAD_MAPPINGS = {'wrong_physical_network': 'wrong_if'}
    BAD_CONFIGS = {'adapter_mappings': BAD_MAPPINGS}

    AGENT = {'alive': True,
             'configurations': GOOD_CONFIGS,
             'host': 'host'}
    AGENTS = [AGENT]

    AGENTS_DEAD = [{'alive': False,
                    'configurations': GOOD_CONFIGS,
                    'host': 'dead_host'}]
    AGENTS_BAD = [{'alive': False,
                   'configurations': GOOD_CONFIGS,
                   'host': 'bad_host_1'},
                  {'alive': True,
                   'configurations': BAD_CONFIGS,
                   'host': 'bad_host_2'}]

    AGENTS_NO_VSWITCH = [{'alive': True,
                          'configurations': NO_VSWITCH_CONFIGS,
                          'host': 'host'}]

    def setUp(self):
        super(DPMMechanismBaseTestCase, self).setUp()
        self.driver = mech_dpm.DPMMechanismDriver()
        self.driver.initialize()


class DPMMechanismGenericTestCase(DPMMechanismBaseTestCase,
                                  base.AgentMechanismGenericTestCase):
    pass


class DPMMechanismFlatTestCase(DPMMechanismBaseTestCase,
                               base.AgentMechanismFlatTestCase):

    def test_type_flat_vif_details(self):
        context = base.FakePortContext(self.AGENT_TYPE,
                                       self.AGENTS,
                                       self.FLAT_SEGMENTS,
                                       vnic_type=self.VNIC_TYPE)
        self.driver.bind_port(context)
        vif_details = context._bound_vif_details
        self.assertIsNone(vif_details.get('vlan'))
        self.assertEqual("fake_vswitch1",
                         vif_details.get('object_id'))

    def test_no_vswitch_ids(self):

        context = base.FakePortContext(self.AGENT_TYPE,
                                       self.AGENTS_NO_VSWITCH,
                                       segments=self.FLAT_SEGMENTS,
                                       vnic_type=self.VNIC_TYPE)
        self.driver.bind_port(context)
        self._check_unbound(context)



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from oslo_log import log

from neutron._i18n import _LW
from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent

LOG = log.getLogger(__name__)

AGENT_TYPE_DPM = 'DPM agent'
VIF_TYPE_DPM_VSWITCH = 'dpm_vswitch'
VIF_TYPE_DPM_ADAPTER = 'dpm_adapter'
VIF_DETAILS_OBJECT_ID = 'object_id'
VLAN_MODE_INBAND = 'inband'


class DPMMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using DPM L2 Agent.

    The DPMMechanismDriver integrates the ml2 plugin with the
    DPM L2 agent. Port binding with this driver requires the
    dpm agent to be running on the port's host, and that agent
    to have connectivity to at least one segment of the port's
    network.
    """

    def __init__(self):
        super(DPMMechanismDriver, self).__init__(
            AGENT_TYPE_DPM,
            [VIF_TYPE_DPM_VSWITCH, VIF_TYPE_DPM_ADAPTER],
            {portbindings.CAP_PORT_FILTER: False})

    def get_allowed_network_types(self, agent):
        return [p_constants.TYPE_FLAT]

    def get_mappings(self, agent):
        return agent.get('configurations', {}).get('adapter_mappings', {})

    def check_vlan_transparency(self, context):
        """DPM driver vlan transparency support."""
        return False

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        if not self.check_segment_for_agent(segment, agent):
            return False
        physnet = segment['physical_network']

        object_ids = self.get_mappings(agent)[physnet]
        if not object_ids:
            LOG.warning(_LW("No Object-IDs found in agents %(agent)s mapping "
                            "for physical network %(net)s."), {'agent': agent,
                                                               'net': physnet})
            return False
        if len(object_ids) > 1:
            LOG.warning(_LW("More than one Object-ID for physical network "
                            "%(net)s on agent %(agent)s found but only a "
                            "single Object-ID is supported. Therefore the "
                            "first one is being chosen!"), {'agent': agent,
                                                            'net': physnet})
        # TODO(andreas_s): In the first release only a single adapter/vswitch
        # per physical network is supported
        object_id = object_ids[0]

        vif_details_segment = self.vif_details

        vif_details_segment[VIF_DETAILS_OBJECT_ID] = object_id
        # TODO(andreas_s): For RoCE Support add the port-element-id to the
        # vif_details
        LOG.debug("DPM vif_details added to context binding: %s",
                  vif_details_segment)
        # TODO(andreas_s) For RoCE Support VIF_TYPE_DPM_ADAPTER is required
        context.set_binding(segment[api.ID], VIF_TYPE_DPM_VSWITCH,
                            vif_details_segment)
        return True



# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_dpm.conf.types import MAPPING_REGEX
from networking_dpm.conf.types import NetworkAdapterMappingType
from networking_dpm.tests import base

VALID_DPM_OBJECT_ID = "fa1f2466-12df-311a-804c-4ed2cc1d656b"
VALID_DPM_OBJECT_ID_UC = "FA1F2466-12DF-311A-804C-4ED2CC1D656B"
VALID_NETWORK_MAPPING = "physnet:" + VALID_DPM_OBJECT_ID + ":0"
VALID_NETWORK_MAPPING_NO_PORT1 = "physnet:" + VALID_DPM_OBJECT_ID + ":"
VALID_NETWORK_MAPPING_NO_PORT2 = "physnet:" + VALID_DPM_OBJECT_ID


class TestNetworkAdapterMappingType(base.BaseTestCase):
    conf_type = NetworkAdapterMappingType()

    def test_valid_mapping(self):
        net, adapter_id, port = self.conf_type(VALID_NETWORK_MAPPING)
        self.assertEqual("physnet", net)
        self.assertEqual("0", port)
        self.assertEqual(VALID_DPM_OBJECT_ID, adapter_id)

    def test_upper_case_to_lower_case_adapter_id(self):
        net, adapter_id, port = self.conf_type(
            "physnet:" + VALID_DPM_OBJECT_ID_UC + ":0")
        self.assertEqual(VALID_DPM_OBJECT_ID, adapter_id)

    def test_keep_case_physnet(self):
        net, adapter_id, port = self.conf_type(
            "PHysnet:" + VALID_DPM_OBJECT_ID_UC + ":0")
        self.assertEqual("PHysnet", net)

    def test_missing_port_default_1_ok(self):
        net, adapter_id, port = self.conf_type(VALID_NETWORK_MAPPING_NO_PORT1)
        self.assertEqual("physnet", net)
        self.assertEqual("0", port)
        self.assertEqual(VALID_DPM_OBJECT_ID, adapter_id)

    def test_missing_port_default_2_ok(self):
        net, adapter_id, port = self.conf_type(VALID_NETWORK_MAPPING_NO_PORT2)
        self.assertEqual("physnet", net)
        self.assertEqual("0", port)
        self.assertEqual(VALID_DPM_OBJECT_ID, adapter_id)

    def test_empty_value_fail(self):
        self.assertRaises(ValueError, self.conf_type, '')

    def test_invalid_value_fail(self):
        self.assertRaises(ValueError, self.conf_type, 'foobar')

    def test_invalid_port_fail(self):
        self.assertRaises(ValueError, self.conf_type,
                          VALID_NETWORK_MAPPING_NO_PORT1 + "2")

    def test_invalid_port_type_fail(self):
        self.assertRaises(ValueError, self.conf_type,
                          VALID_NETWORK_MAPPING_NO_PORT1 + "a")

    def test_invalid_adapter_id_fail(self):
        self.assertRaises(ValueError, self.conf_type, "physnet:foo:1")

    def test_invalid_pysnet_fail(self):
        invalid = ["phys:net:" + VALID_DPM_OBJECT_ID,
                   "phys:net:" + VALID_DPM_OBJECT_ID + ":",
                   "phys:net:" + VALID_DPM_OBJECT_ID + ":1",
                   "foo:phys:net:" + VALID_DPM_OBJECT_ID,
                   ":net:" + VALID_DPM_OBJECT_ID,
                   "::net:" + VALID_DPM_OBJECT_ID]
        for conf in invalid:
            self.assertRaises(ValueError, self.conf_type, conf)

    def test_repr(self):
        self.assertEqual(
            "String(regex='" + MAPPING_REGEX + "')",
            repr(self.conf_type))

    def test_equal(self):
        self.assertTrue(
            NetworkAdapterMappingType() == NetworkAdapterMappingType())



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_dpm.tests import base
from networking_dpm.tests.unit import fake_zhmcclient


class TestFakeZHMC(base.BaseTestCase):

    def test_fakehmc_full(self):

        nics = [{'name': 'nic-name-1', 'object-id': 'nic-id-1'},
                {'name': 'nic-name-2', 'object-id': 'nic-id-2'}]
        vswitches = [{'backing-adapter-uri': '/api/adapters/uuid-0',
                      'object-id': 'vswitch-uuid-0', 'port': 0},
                     {'backing-adapter-uri': '/api/adapters/uuid-1',
                      'object-id': 'vswitch-uuid-1',
                      'port': 1,
                      'nics': nics}
                     ]
        adapter_ports = [{'element-id': '0'}, {'element-id': '1'}]
        adapters = [{'object-id': 'uuid-0', 'ports': adapter_ports},
                    {'object-id': 'uuid-1'}]

        hmc_json = {'cpcs': [
            {'object-id': 'oid0', 'name': 'CPCname0',
             'vswitches': vswitches, 'adapters': adapters},
            {'object-id': 'oid1', 'name': 'CPCname1', 'dpm_enabled': False,
             'vswitches': {}}
        ]}

        client = fake_zhmcclient.get_client(hmc_json)
        cpc_list = client.cpcs.list()
        self.assertTrue(len(cpc_list) == 2)

        # CPC 0
        cpc0 = client.cpcs.find(**{'object-id': 'oid0'})
        self.assertEqual('oid0', cpc0.get_property('object-id'))
        self.assertEqual('CPCname0', cpc0.get_property('name'))
        self.assertIn(cpc0, cpc_list)
        self.assertTrue(cpc0.dpm_enabled)

        vswitch_list = cpc0.vswitches.list()
        self.assertEqual(2, len(vswitch_list))

        vswitch0 = cpc0.vswitches.find(**{'object-id': 'vswitch-uuid-0'})
        self.assertEqual('/api/adapters/uuid-0',
                         vswitch0.get_property('backing-adapter-uri'))
        self.assertEqual('vswitch-uuid-0', vswitch0.get_property('object-id'))
        self.assertEqual(0, vswitch0.get_property('port'))
        self.assertIn(vswitch0, vswitch_list)

        vswitch1 = cpc0.vswitches.find(**{'object-id': 'vswitch-uuid-1'})
        self.assertEqual('/api/adapters/uuid-1',
                         vswitch1.get_property('backing-adapter-uri'))
        self.assertEqual('vswitch-uuid-1', vswitch1.get_property('object-id'))
        self.assertEqual(1, vswitch1.get_property('port'))
        self.assertIn(vswitch1, vswitch_list)

        adapter_list = cpc0.adapters.list()
        self.assertEqual(2, len(adapter_list))
        adapter0 = cpc0.adapters.find(**{'object-id': 'uuid-0'})
        self.assertEqual('uuid-0', adapter0.get_property('object-id'))
        port_list = adapter0.ports.list()
        self.assertEqual(2, len(port_list))
        port0 = adapter0.ports.find(**{'element-id': '0'})
        self.assertEqual('0', port0.get_property('element-id'))
        port1 = adapter0.ports.find(**{'element-id': '1'})
        self.assertEqual('1', port1.get_property('element-id'))

        adapter1 = cpc0.adapters.find(**{'object-id': 'uuid-1'})
        self.assertEqual('uuid-1', adapter1.get_property('object-id'))

        nics = vswitch1.get_connected_nics()
        self.assertEqual(2, len(nics))
        nic1 = nics[0]
        self.assertEqual(('nic-name-1', 'nic-id-1'),
                         (nic1.get_property('name'),
                          nic1.get_property('object-id')))
        nic2 = nics[1]
        self.assertEqual(('nic-name-2', 'nic-id-2'),
                         (nic2.get_property('name'),
                          nic2.get_property('object-id')))

        # CPC 1
        cpc1 = client.cpcs.find(**{'object-id': 'oid1'})
        self.assertEqual('oid1', cpc1.get_property('object-id'))
        self.assertEqual('CPCname1', cpc1.get_property('name'))
        self.assertIn(cpc1, cpc_list)
        self.assertFalse(cpc1.dpm_enabled)

        self.assertEqual(0, len(cpc1.vswitches.list()))

    def test_port_element_id_int_fail(self):
        """zhmcclient treats element-id as string"""
        port_json = {'element-id': 0}
        try:
            fake_zhmcclient._Port(port_json)
            self.fail("Type error should be thrown, as port-element-id was "
                      "provided as int")
        except TypeError:
            pass

    def test_port_element_id_string_ok(self):
        """zhmcclient treats element-id as string"""
        port_json = {'element-id': '0'}
        self.assertIsNotNone(fake_zhmcclient._Port(port_json))

    def test_vswitch_port_string_fail(self):
        """zhmcclient treats vswitch 'port' attribute as int

        But 'port-element-id' of the adapter is treated as String
        """
        vswitches_json = {'port': '0'}
        try:
            fake_zhmcclient._VSwitch(vswitches_json)
            self.fail("Type error should be thrown, as port was provided as "
                      "int")
        except TypeError:
            pass

    def test_vswitch_port_int_ok(self):
        """zhmcclient treats vswitch 'port' attribute as int

        But 'port-element-id' of the adapter is treated as String
        """
        vswitches_json = {'port': 0}

        self.assertIsNotNone(fake_zhmcclient._VSwitch(vswitches_json))

    def test_fakehmc_consistency_fail(self):
        vswitches = [{'backing-adapter-uri': '/api/adapters/uuid-0',
                      'object-id': 'vswitch-uuid-0'}]
        adapters = [{'object-id': 'another-uuid'}]
        hmc_json = {'cpcs': [
            {'object-id': 'oid0',
             'vswitches': vswitches, 'adapters': adapters}
        ]}
        self.assertRaises(fake_zhmcclient.FakeZHMCClientError,
                          fake_zhmcclient.get_cpc, hmc_json)

    def test_fakehmc_consistency_given(self):
        vswitches = [{'backing-adapter-uri': '/api/adapters/uuid-0',
                      'object-id': 'vswitch-uuid-0'}]
        adapters = [{'object-id': 'uuid-0'}]
        hmc_json = {'cpcs': [
            {'object-id': 'oid0',
             'vswitches': vswitches, 'adapters': adapters}
        ]}
        # No Exception thrown
        fake_zhmcclient.get_cpc(hmc_json)

    def test_fakehmc_consistency_no_adapters_given(self):
        # If no adapters are specified, don't check for consistency
        vswitches = [{'backing-adapter-uri': '/api/adapters/uuid-0',
                      'object-id': 'vswitch-uuid-0'}]
        hmc_json = {'cpcs': [
            {'object-id': 'oid0', 'vswitches': vswitches}
        ]}
        # No Exception thrown
        fake_zhmcclient.get_cpc(hmc_json)

    def test_fakehmc_minimal(self):
        hmc = {"cpcs": [{"name": "foo"}]}
        # Verify that no exception is thrown
        fake_zhmcclient.get_client(hmc)

    def test_fakehmc_str(self):
        test_json = '{"cpcs": [{"prop1": "value1", "vswitches": []}]}'
        # Verify that no exception is thrown
        fake_zhmcclient.get_client(test_json)

    def test_fakehmc_invalid_json(self):
        test_json = "{'blub': {"
        self.assertRaises(ValueError, fake_zhmcclient.get_client, test_json)

    def test_fakehmc_get_cpc(self):
        hmc = {"cpcs": [{"name": "foo"}]}
        cpc = fake_zhmcclient.get_cpc(hmc)
        self.assertEqual("foo", cpc.get_property('name'))



# Copyright 2017 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from oslo_config import cfg
from oslo_config.fixture import Config
import tempfile

from networking_dpm.conf.cfg import MultiNetworkAdapterMappingOpt
from networking_dpm.conf.types import NetworkAdapterMappingType
from networking_dpm.tests import base


class TestNetworkAdapterMappingType(base.BaseTestCase):

    def create_tempfiles(self, files, ext='.conf'):
        """Create temp files for testing

        :param files: A list of files of tuples in the format
           [('filename1', 'line1\nline2\n'), ('filename2', 'line1\nline2\n')]
        :param ext: The file extension to be used
        :return: List of file paths
           paths[0] = path of filename1
           paths[1] = path of filename2
        """
        # TODO(andreas_s): Make a mixin in os-dpm and move this there
        # (also required in nova-dpm)
        tempfiles = []
        for (basename, contents) in files:
            if not os.path.isabs(basename):
                # create all the tempfiles in a tempdir
                tmpdir = tempfile.mkdtemp()
                path = os.path.join(tmpdir, basename + ext)
                # the path can start with a subdirectory so create
                # it if it doesn't exist yet
                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))
            else:
                path = basename + ext
            fd = os.open(path, os.O_CREAT | os.O_WRONLY)
            tempfiles.append(path)
            try:
                os.write(fd, contents.encode('utf-8'))
            finally:
                os.close(fd)
        return tempfiles

    def test_object(self):
        opt = MultiNetworkAdapterMappingOpt("mapping", help="foo-help")
        self.assertEqual("foo-help", opt.help)
        self.assertEqual("mapping", opt.name)
        self.assertEqual(NetworkAdapterMappingType, type(opt.type))

    def test_config_single_set_override(self):
        mapping = ["physnet:aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa:0"]
        opt = MultiNetworkAdapterMappingOpt("mapping")
        cfg.CONF.register_opt(opt)
        self.flags(mapping=mapping)
        self.assertEqual([("physnet", "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                           "0")], cfg.CONF.mapping)

    def test_config_multiple_set_override(self):
        mapping = ["physnet:aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa:0",
                   "net2:bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"]
        opt = MultiNetworkAdapterMappingOpt("mapping")
        cfg.CONF.register_opt(opt)
        self.flags(mapping=mapping)
        expected_mapping = [
            ("physnet", "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "0"),
            ("net2", "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb", "0")]
        self.assertEqual(expected_mapping, cfg.CONF.mapping)

    def test_config_multiple_with_conf_file(self):
        mapping = "physnet:aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa:0"
        other_mapping = "net2:bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
        paths = self.create_tempfiles([
            ('test', '[DEFAULT]\nmapping = ' + mapping + '\n'
             'mapping = ' + other_mapping + '\n')])

        conf = self.useFixture(Config())
        conf.register_opt(MultiNetworkAdapterMappingOpt("mapping"))
        conf.conf(args=['--config-file', paths[0]])
        expected_mapping = [
            ("physnet", "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "0"),
            ("net2", "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb", "0")]
        self.assertEqual(expected_mapping, cfg.CONF.mapping)
